import socket
import threading
from typing import Tuple
from typing import Type
from typing import TypeAlias
from typing import Union

from msu_ssc import ssc_log
from msu_ssc.time_util import utc
from msu_ssc.udp_mux import _shutdown_socket
from msu_ssc.udp_mux import _tup_to_str

IPv4SockTup: TypeAlias = Tuple[str, int]


class OneWayUdpProxyThread(threading.Thread):
    def __init__(
        self,
        *,
        source_tup: IPv4SockTup,
        destination_tup: IPv4SockTup,
        proxy_tup: IPv4SockTup,
        daemon: bool = True,
        name: str = "proxy",
        **kwargs,
    ):
        name += f"({_tup_to_str(proxy_tup)}->{_tup_to_str(destination_tup)})"

        super().__init__(
            name=name,
            daemon=daemon,
            **kwargs,
        )
        self.source_tup = source_tup
        self.destination_tup = destination_tup
        self.proxy_tup = proxy_tup
        self.proxy_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self.total_packets = 0
        self.total_bytes = 0

    # def target

    def run(self):
        # BIND
        ssc_log.info(f"Attempting to bind to UDP socket {_tup_to_str(self.proxy_tup)} for receiving. [{self.name}]")
        _shutdown_socket(self.proxy_socket)
        self.proxy_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.proxy_socket.bind(self.proxy_tup)
        ssc_log.info(f"Successfully bound receiving socket {_tup_to_str(self.proxy_tup)}. [{self.name}]")

        # SERVE FOREVER
        self._mux_start_time = utc()
        ssc_log.info(
            f"Ready to begin proxying at {self._mux_start_time.isoformat(timespec='seconds', sep=' ')}. [{self.name}]"
        )
        while True:
            data, source_address = self.proxy_socket.recvfrom(4096)
            self._receive_packet(
                data=data,
                source_address=source_address,
            )

    def handle_packet(
        self,
        *,
        data: bytes,
        destination_tup: IPv4SockTup,
        **kwargs,
    ):
        """Overload this one."""
        ssc_log.debug(f"sending {len(data)} to {_tup_to_str(destination_tup)} [{self.name}]")
        self.proxy_socket.sendto(data, destination_tup)

    def _receive_packet(
        self,
        *,
        data: bytes,
        source_address: Union[IPv4SockTup, None] = None,
        debug: bool = True,
    ) -> None:
        if debug:
            message = f"Received {len(data)} bytes"
            if source_address:
                message += f" from {_tup_to_str(source_address)}."
            else:
                message += "."
            message += f" (total: {self.total_packets} packets; {self.total_bytes} bytes) "
            message += f"[{self.name}]"
            ssc_log.debug(message)

        self.handle_packet(
            data=data,
            destination_tup=self.destination_tup,
        )
        self.total_bytes += len(data)
        self.total_packets += 1


class BidirectionalUdpProxy:
    thread_class: Type[OneWayUdpProxyThread] = OneWayUdpProxyThread

    def __init__(
        self,
        *,
        server_tup: Tuple[str, int],
        client_tup: Tuple[str, int],
        server_proxy_tup: Tuple[str, int],
        client_proxy_tup: Tuple[str, int],
    ):
        self.server_tup = server_tup
        self.client_tup = client_tup
        self.server_proxy_tup = server_proxy_tup
        self.client_proxy_tup = client_proxy_tup

        self.server_to_client = self.__class__.thread_class(
            daemon=True,
            source_tup=self.server_tup,
            destination_tup=self.client_tup,
            proxy_tup=self.client_proxy_tup,
            name="server_to_client",
        )
        self.client_to_server = self.__class__.thread_class(
            daemon=True,
            source_tup=self.client_tup,
            destination_tup=self.server_tup,
            proxy_tup=self.server_proxy_tup,
            name="client_to_server",
        )

        self.server_to_client.start()
        self.client_to_server.start()

        ssc_log.debug(f"self.server_to_client={self.server_to_client!r} type={type(self.server_to_client)!r}")
        ssc_log.debug(f"self.client_to_server={self.client_to_server!r} type={type(self.client_to_server)!r}")

    def __repr__(self):
        return f"{self.__class__.__name__}({_tup_to_str(self.server_tup)}<->{_tup_to_str(self.client_tup)})"


class OneWayUdpProxyThreadFailure(OneWayUdpProxyThread):
    """A proxy thread where every other packet will die"""

    def handle_packet(self, *, data: bytes, destination_tup: IPv4SockTup, **kwargs):
        if self.total_packets % 2 == 0:
            ssc_log.info(f"INTENTIONAL FAILURE. packet index: {self.total_packets} [{self.name}]")
        else:
            ssc_log.debug(f"Sending packet normally [{self.name}]")
            self.proxy_socket.sendto(data, destination_tup)


class BidirectionalUdpProxyFailure(BidirectionalUdpProxy):
    thread_class = OneWayUdpProxyThreadFailure


if __name__ == "__main__":
    import time

    proxy_class = BidirectionalUdpProxy
    # proxy_class = BidirectionalUdpProxyFailure

    proxy = proxy_class(
        server_tup=("127.0.0.1", 8003),
        client_tup=("127.0.0.1", 8002),
        server_proxy_tup=("127.0.0.1", 9003),
        client_proxy_tup=("127.0.0.1", 9002),
    )
    ssc_log.info(proxy)
    sleep_secs = 1000000
    ssc_log.debug(f"Sleeping {sleep_secs:,} secs")
    time.sleep(sleep_secs)
    ssc_log.warning(f"Done sleeping. Exiting")



import datetime

import freezegun

from msu_ssc import ssc_log


def test_utc_filename_timestamp():
    with freezegun.freeze_time(
        "2025-01-02 03:45:56.123456-05:00",
        tz_offset=datetime.timedelta(seconds=-5 * 3600),  # -05:00, EST
    ):
        assert ssc_log.utc_filename_timestamp() == "2025-01-02T03_45_56.log"

        # TIMESPECS
        assert ssc_log.utc_filename_timestamp(timespec="auto") == "2025-01-02T03_45_56.123456.log"
        assert ssc_log.utc_filename_timestamp(timespec="hours") == "2025-01-02T03.log"
        assert ssc_log.utc_filename_timestamp(timespec="minutes") == "2025-01-02T03_45.log"
        assert ssc_log.utc_filename_timestamp(timespec="seconds") == "2025-01-02T03_45_56.log"
        assert ssc_log.utc_filename_timestamp(timespec="milliseconds") == "2025-01-02T03_45_56.123.log"
        assert ssc_log.utc_filename_timestamp(timespec="microseconds") == "2025-01-02T03_45_56.123456.log"

        # PREFIX
        assert ssc_log.utc_filename_timestamp(prefix="prefix") == "prefix_2025-01-02T03_45_56.log"

        # SUFFIX
        assert ssc_log.utc_filename_timestamp(suffix="suffix") == "2025-01-02T03_45_56_suffix.log"

        # EXTENSION
        assert ssc_log.utc_filename_timestamp(extension=".txt") == "2025-01-02T03_45_56.txt"
        assert ssc_log.utc_filename_timestamp(extension="txt") == "2025-01-02T03_45_56.txt"

        # PREFIX, SUFFIX, EXTENSION
        assert (
            ssc_log.utc_filename_timestamp(
                prefix="prefix",
                suffix="suffix",
                extension=".txt",
            )
            == "prefix_2025-01-02T03_45_56_suffix.txt"
        )



import datetime
import logging
import sys
from pathlib import Path
from typing import Union

if sys.version_info >= (3, 8):
    from typing import Literal

    TimespecType = Literal["auto", "hours", "minutes", "seconds", "milliseconds", "microseconds"]
else:
    TimespecType = str

from msu_ssc.path_util import clean_path_part
from msu_ssc.path_util import file_timestamp

try:
    from rich.logging import RichHandler

    _rich_imported = True
except ImportError:
    _rich_imported = False

logger = logging.getLogger("ssc")
"""The primary logger. It will have the name `ssc`. Use `ssc_log.init()` to configure."""


def utc_filename_timestamp(
    timestamp: Union[datetime.datetime, None] = None,
    *,
    prefix: str = "",
    suffix: str = "",
    extension: str = ".log",
    timespec: TimespecType = "seconds",
    assume_utc=False,
    assume_local=False,
) -> str:
    """Generate a filename with a UTC timestamp.

    With all defaults, will give a filename like `2025-02-03T12_34_56.log`

    To make a `equipment_2025-02-03T12_34_56.789_log.txt`, you would call:

    ```
    filename = utc_filename_timestamp(
        prefix="equipment",
        suffix="log",
        extension=".txt",
        timespec="milliseconds",
    )
    ```
    """
    timestamp_string = file_timestamp(
        timestamp=timestamp,
        sep="T",
        timespec=timespec,
        desired_tz=datetime.timezone.utc,
        assume_utc=assume_utc,
        assume_local=assume_local,
    )
    if not extension.startswith("."):
        extension = "." + extension
    return clean_path_part("_".join(x for x in (prefix, timestamp_string, suffix) if x) + extension)


# logger.setLevel("DEBUG")

plain_text_formatter = logging.Formatter(
    fmt="[%(asctime)s.%(msecs)03d %(levelname)-8s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def _iso_str(timestamp: "datetime.datetime") -> str:
    string = f"{timestamp:%Y-%m-%d %H:%M:%S.%f}"
    milliseconds_string = string[:-3]
    return f"[{milliseconds_string}]"


if _rich_imported:
    console_handler = RichHandler(
        level="DEBUG",
        show_time=True,
        omit_repeated_times=False,
        rich_tracebacks=True,
        # log_time_format="[%Y-%m-%d %H:%M:%S.%f]",
        log_time_format=_iso_str,
    )
else:
    console_handler = logging.StreamHandler()
    console_handler.setLevel("DEBUG")
    console_handler.setFormatter(plain_text_formatter)


def init(
    level: Union[str, None] = "INFO",
    *,
    plain_text_file_path: Union[Path, str, None] = None,
    jsonl_file_path: Union[Path, str, None] = None,
    plain_text_level: Union[str, None] = None,
    jsonl_level: Union[str, None] = None,
    console_level: Union[str, None] = None,
) -> None:
    if level:
        logger.setLevel(level.upper())

    if plain_text_file_path:
        plain_text_level = plain_text_level or level
        _log_to_file(
            plain_text_file_path,
            level=plain_text_level,
        )

    if jsonl_file_path:
        jsonl_level = jsonl_level or level
        _log_to_jsonl_file(
            jsonl_file_path,
            level=jsonl_level,
        )

    console_level = console_level or level
    console_handler.setLevel(console_level)
    logger.addHandler(console_handler)


DEFAULT_LOG_DIRECTORY = Path(__file__).expanduser().resolve().parent.parent / "logs"
"""Should be `./logs/`"""


def log_to_default_file() -> None:
    """Begin logging to default file, which will be a timestamped file in ./logs/

    Name will be like `./logs/stk_2025-02-03T12_34_56.log`
    """
    import datetime

    file_name = f"stk_{datetime.datetime.now():%Y-%m-%dT%H_%M_%S}.log"
    file_path = DEFAULT_LOG_DIRECTORY / file_name
    _log_to_file(path=file_path)


def _log_to_file(
    path: Union[Path, str],
    level: Union[str, None] = None,
    *,
    encoding: str | None = "utf-8",
) -> None:
    """Begin logging in plaintext to the given file. File will be APPENDED, and encoded in UTF-8"""
    resolved_path = Path(path).expanduser().resolve()
    if not resolved_path.parent.exists():
        resolved_path.parent.mkdir(
            parents=True,
            exist_ok=True,
        )

    file_handler = logging.FileHandler(
        filename=Path(resolved_path).expanduser().resolve(),
        encoding=encoding,
    )

    if level:
        file_handler.setLevel(level)
    file_handler.setFormatter(plain_text_formatter)
    logger.addHandler(file_handler)
    logger.debug(f"Begin logging to {resolved_path.__fspath__()!r}")


def _log_to_jsonl_file(
    path: Union[Path, str],
    level: Union[str, None] = None,
) -> None:
    """Begin logging to the given file. File will be APPENDED, and encoded in UTF-8"""
    try:
        from pythonjsonlogger.jsonlogger import JsonFormatter
        # from pythonjsonlogger.json import JsonFormatter
    except ImportError:
        logger.error("pythonjsonlogger is not installed. Please install it to use JSON logging.")
        raise

    resolved_path = Path(path).expanduser().resolve()
    if not resolved_path.parent.exists():
        resolved_path.parent.mkdir(
            parents=True,
            exist_ok=True,
        )
    file_handler = logging.FileHandler(filename=Path(resolved_path).expanduser().resolve())

    if level:
        file_handler.setLevel(level)
    json_formatter = JsonFormatter(
        reserved_attrs=(
            "msg",
            "args",
            "levelno",
        ),
    )
    file_handler.setFormatter(json_formatter)
    logger.addHandler(file_handler)
    logger.debug(f"Begin logging to {resolved_path.__fspath__()!r}")


def getChild(name):
    return logger.getChild(name)


log = logger.log
debug = logger.debug
info = logger.info
warning = logger.warning
error = logger.error
critical = logger.critical


if __name__ == "__main__":
    # log_to_file(r"C:\Users\msu\Desktop\LEMS - InProgress\Analysis\logs\log.log")
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    init(
        level="DEBUG",
        # plain_text_file_path="logs/log.log",
        plain_text_file_path=f"logs/{utc_filename_timestamp(now, prefix=logger.name, extension='.log')}",
        jsonl_file_path=f"logs/{utc_filename_timestamp(now, prefix=logger.name, extension='.jsonl')}",
    )
    # log_to_default_file()

    debug("DEBUG MESSAGE")
    info("INFO MESSAGE")
    warning("WARNING MESSAGE")
    error("ERROR MESSAGE")
    critical("CRITICAL MESSAGE")

    # info("\n\nThis has\n\n\n\multiple newlines\n\n\n")

    # try:
    #     1 / 0
    # except Exception as exc:
    #     error("An error occurred", exc_info=exc)

    child = logger.getChild("child")

    child.info("Child info")



import datetime
import socket
import threading
from typing import Iterable
from typing import Tuple
from typing import Union

from msu_ssc import ssc_log

# logger = create_logger(__file__, level="DEBUG")


def _tup_to_str(socket_tuple: Tuple[str, int]) -> str:
    return f"{socket_tuple[0]}:{socket_tuple[1]}"


def _str_to_tup(socket_str: str) -> Tuple[str, int]:
    host_str, port_str = socket_str.strip().split(":")
    return (host_str, int(port_str))


def _shutdown_socket(sock: socket.socket):
    try:
        sock.close()
    except Exception as exc:
        ssc_log.warning(f"Unable to shutdown socket {sock}", exc_info=exc)


class UdpMux:
    def __init__(
        self,
        receive_socket_tuple: Tuple[str, int],
        transmit_socket_tuples: Union[Iterable[Tuple[str, int]], None] = None,
        *,
        daemon=True,
        reuse_receive_socket: bool = False,
    ) -> None:
        self.receive_socket_tuple = receive_socket_tuple
        self.transmit_socket_tuples = list(transmit_socket_tuples or [])
        self.daemon = daemon
        self.reuse_receive_socket = reuse_receive_socket

        self.receive_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.transmit_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self._bound = False
        self._received_packet_count = 0
        self._received_bytes_count = 0
        self._transmitted_packet_count = 0
        self._transmitted_bytes_count = 0

        self.thread = threading.Thread(
            name=f"udp-mux-{_tup_to_str(self.receive_socket_tuple)}",
            daemon=self.daemon,
            target=self.start_mux,
        )

        self.thread.start()

    def start_mux(self) -> None:
        ssc_log.info(f"Beginning MUX setup")
        self.bind()
        self._mux_start_time = datetime.datetime.now(tz=datetime.timezone.utc)
        ssc_log.info(f"Ready to begin muxing at {self._mux_start_time.isoformat(timespec='seconds', sep=' ')}.")
        while True:
            data, source_address = self.receive_socket.recvfrom(4096)
            self.handle_packet(data, source_address)

    def stop_mux(self) -> None:
        self._mux_stop_time = datetime.datetime.now(tz=datetime.timezone.utc)
        try:
            elapsed = (self._mux_stop_time - self._mux_start_time).total_seconds()
        except Exception:
            elapsed = 0
        ssc_log.info(
            f"Stopping muxing at {self._mux_stop_time.isoformat(timespec='seconds', sep=' ')}. Muxed for {elapsed:.2f} seconds ({elapsed / 3600:.4f} hours)."
        )
        _shutdown_socket(self.receive_socket)
        if self.receive_socket is not self.transmit_socket:
            _shutdown_socket(self.transmit_socket)
        ssc_log.debug(
            f"Received {self._received_packet_count} packets ({self._received_bytes_count} bytes). "
            + f"Transmitted {self._transmitted_packet_count} packets ({self._transmitted_bytes_count} bytes)."
        )

    def bind(self) -> None:
        # RECEIVE
        _shutdown_socket(self.receive_socket)
        self.receive_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        ssc_log.info(f"Attempting to bind to UDP socket {_tup_to_str(self.receive_socket_tuple)} for receiving.")
        self.receive_socket.bind(self.receive_socket_tuple)
        ssc_log.info(f"Successfully bound receiving socket.")

        # TRANSMIT
        _shutdown_socket(self.transmit_socket)
        if self.reuse_receive_socket:
            ssc_log.debug(
                f"Reusing receiving UDP socket {_tup_to_str(self.receive_socket_tuple)} as transmitting socket."
            )
            self.transmit_socket = self.receive_socket
        else:
            send_socket_tuple = self.transmit_socket_tuples[0]
            ssc_log.info(f"Attempting to bind to UDP socket {_tup_to_str(send_socket_tuple)} for transmitting.")
            self.receive_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            ssc_log.info(f"Successfully bound transmitting socket.")
        self._bound = True

    def handle_packet(self, payload_data: bytes, source_address=None) -> None:
        self._received_packet_count += 1
        self._received_bytes_count += len(payload_data)
        ssc_log.debug(f"Received {len(payload_data):,} bytes from {_tup_to_str(source_address)}")
        for transmit_socket_tuple in self.transmit_socket_tuples:
            attempted_transmitted_data_size = len(payload_data)
            ssc_log.debug(
                f"  Sending {attempted_transmitted_data_size:,} bytes to {_tup_to_str(transmit_socket_tuple)}"
            )
            actual_transmitted_data_size = self.transmit_socket.sendto(payload_data, transmit_socket_tuple)
            if actual_transmitted_data_size != attempted_transmitted_data_size:
                ssc_log.error(
                    (
                        f"Error transmitting packet, originally from {source_address}, to {transmit_socket_tuple}. "
                        + f"Attempted to send {attempted_transmitted_data_size} bytes, actually sent {actual_transmitted_data_size} bytes."
                    ),
                    extra={
                        "payload": {payload_data},
                    },
                )
            self._transmitted_packet_count += 1
            self._transmitted_bytes_count += actual_transmitted_data_size

    def __enter__(self) -> "UdpMux":
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop_mux()


def main() -> int:
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "receive",
        action="store",
        help="UDP socket to receive on",
    )
    parser.add_argument(
        "--transmit",
        "-T",
        nargs="*",
        help="UDP sockets to retransmit on. Give as a list of separated sockets, like `-T 127.0.0.1:8001 127.0.0.1:8002`",
        default=(),
    )
    parser.add_argument(
        "--log-level",
        "-L",
        help="Console log level",
        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        default="INFO",
        action="store",
    )
    parser.add_argument(
        "--reuse-socket",
        action="store_true",
        help=(
            "Reuse the socket that receives the segments when retransmitting them. "
            + "This will cause the source port of the retransmitted packets to be the same as if the packet never passed through this muxer.",
        ),
    )
    args = parser.parse_args()
    receive_socket_tuple = _str_to_tup(args.receive)
    transmit_socket_tuples = [_str_to_tup(sock_str) for sock_str in args.transmit]
    ssc_log.debug(f"Parsed command line arguments: {args!r}")
    ssc_log.debug(f"Parsed receive UDP socket: {receive_socket_tuple!r}")
    ssc_log.debug(f"Parsed transmit UCP socket(s): {transmit_socket_tuples!r}")

    with UdpMux(
        receive_socket_tuple=receive_socket_tuple,
        transmit_socket_tuples=transmit_socket_tuples,
        reuse_receive_socket=args.reuse_socket,
        daemon=True,
        log_level=args.log_level,
    ) as mux:  # noqa: F841
        import time

        time.sleep(5)
        pass

    return 0


if __name__ == "__main__":
    import sys

    sys.exit(main())



import datetime
import re
import sys
from pathlib import Path
from typing import Union

if sys.version_info >= (3, 8):
    from typing import Literal

    TimespecType = Literal["auto", "hours", "minutes", "seconds", "milliseconds", "microseconds"]
else:
    TimespecType = str

_chunk_regex = re.compile(r"[^a-zA-Z0-9\-_\.]+")


def clean_path_part(part: str) -> str:
    return re.sub(_chunk_regex, "_", part)


def _is_valid_path_chunk(chunk: str) -> bool:
    return chunk == clean_path_part(chunk)


def clean_path(path: Path) -> Path:
    return Path(*(clean_path_part(part) for part in Path(path).parts))


def file_timestamp(
    timestamp: Union[datetime.datetime, None] = None,
    *,
    sep="T",
    timespec: TimespecType = "seconds",
    assume_utc=False,
    assume_local=False,
    desired_tz: Union[datetime.tzinfo, None] = None,
) -> str:
    """
    Convert a timestamp to a string suitable for use in a file name.

    Using all defaults will give you a string like `"2025-01-01T00_00_00"`, where the time is the current system time in UTC.

    TIME ZONES:
    - The `desired_tz` parameter is the time zone to which the timestamp will be converted before formatting. Default is UTC.
    - If the input is naive (i.e., has no time zone info), you must pass either `assume_utc=True` or `assume_local=True`. If
    you want something more complex, you'll need to convert the timestamp yourself before passing it in.

    Args:
        timestamp: The timestamp to convert. If None, the current time is used.
        sep: The separator between the date and time components.
        timespec: The level of precision to include in the time component.
        assume_utc: If True, assume the timestamp is in UTC.
        assume_local: If True, assume the timestamp is in the local time zone.
        desired_tz: The time zone to which the timestamp will be converted before formatting. Default is UTC.

    Returns:
        A string representation of the timestamp.
    """
    timestamp = timestamp or datetime.datetime.now(tz=datetime.timezone.utc)
    desired_tz = desired_tz or datetime.timezone.utc

    naive = timestamp.tzinfo is None
    if naive:
        if assume_utc and assume_local:
            raise ValueError("Cannot assume both UTC and local time")
        elif not assume_utc and not assume_local:
            raise ValueError("Must assume either UTC or local time if timestamp is naive")
        if assume_utc:
            timestamp = timestamp.replace(tzinfo=datetime.timezone.utc)
        elif assume_local:
            timestamp = timestamp.astimezone(datetime.timezone.utc)

    assert timestamp.tzinfo is not None, "Timestamp must have time zone info"

    # This is weird: We need to convert to UTC, but then strip the time zone info
    # to avoid the "+00:00" in the string
    timestamp = timestamp.astimezone(desired_tz).replace(tzinfo=None)

    return clean_path_part(timestamp.isoformat(sep=sep, timespec=timespec))


if __name__ == "__main__":
    pass
    strings = [
        "",
        ".git",
        "2025-01-02T12:34:56.789012",
    ]

    for string in strings:
        print(f"string={string!r}")
        print(f"clean_path_part(string)={clean_path_part(string)!r}")
        print(f"_is_valid_path_chunk(string)={_is_valid_path_chunk(string)!r}")
        print()

    print(clean_path(Path("~/2025-01-02T12:34:56.789012.log")))

    class TZ(datetime.tzinfo):
        """A time zone with an arbitrary, constant -06:39 offset."""

        def utcoffset(self, dt):
            return datetime.timedelta(hours=+6, minutes=+39)

    dt = datetime.datetime(
        2025,
        1,
        2,
        12,
        00,
        00,
        0,
        # tzinfo=TZ(),
    )

    print(
        file_timestamp(
            dt,
            timespec="milliseconds",
            # assume_utc=True,
            assume_local=True,
        )
    )

    print(file_timestamp())



import datetime

from msu_ssc import ssc_log

timestamp_utc = datetime.datetime(2025, 1, 2, 3, 45, 56, 123456, tzinfo=datetime.timezone.utc)
timestamp_naive = datetime.datetime(2025, 1, 2, 3, 45, 56, 123456)


def test_file_timestamp():
    assert ssc_log.file_timestamp(timestamp=timestamp_utc) == "2025-01-02T03_45_56"

    # TIMESPECS
    assert ssc_log.file_timestamp(timestamp=timestamp_utc, timespec="auto") == "2025-01-02T03_45_56.123456"
    assert ssc_log.file_timestamp(timestamp=timestamp_utc, timespec="hours") == "2025-01-02T03"
    assert ssc_log.file_timestamp(timestamp=timestamp_utc, timespec="minutes") == "2025-01-02T03_45"
    assert ssc_log.file_timestamp(timestamp=timestamp_utc, timespec="seconds") == "2025-01-02T03_45_56"
    assert ssc_log.file_timestamp(timestamp=timestamp_utc, timespec="milliseconds") == "2025-01-02T03_45_56.123"
    assert ssc_log.file_timestamp(timestamp=timestamp_utc, timespec="microseconds") == "2025-01-02T03_45_56.123456"

    # ASSUME_UTC
    assert ssc_log.file_timestamp(timestamp=timestamp_naive, assume_utc=True) == "2025-01-02T03_45_56"

    # ASSUME_LOCAL
    # My word, this is annoying.
    # timestamp_naive = datetime.datetime(2025, 1, 15)

    # Get _tzinfo.   https://stackoverflow.com/a/39079819
    local_timezone: datetime.timezone = timestamp_naive.astimezone().tzinfo
    utcoffset = local_timezone.utcoffset(timestamp_naive)
    simple_timestamp_naive_plus_offset = timestamp_naive - utcoffset
    manually_converted_to_utc = simple_timestamp_naive_plus_offset.replace(tzinfo=datetime.timezone.utc)
    automatically_converted_to_utc = timestamp_naive.astimezone(datetime.timezone.utc)

    assert manually_converted_to_utc == automatically_converted_to_utc
    assert ssc_log.file_timestamp(timestamp=timestamp_naive, assume_local=True) == ssc_log.file_timestamp(
        timestamp=manually_converted_to_utc,
    )


def test_utc_filename_timestamp():
    import datetime

    assert ssc_log.utc_filename_timestamp(timestamp=timestamp_utc) == "2025-01-02T03_45_56.log"

    assert datetime
    pass


if __name__ == "__main__":
    # pytest.main([__file__])
    pass



import datetime


def utc() -> datetime.datetime:
    """Get the current system time, in UTC, as a timezone-aware `datetime` object.

    Note that this is not the same as `datetime.utcnow()`, whish returns a naive `datetime`
    object, and is deprecated as of Python 3.12. See https://docs.python.org/3/library/datetime.html#datetime.datetime.utcnow

    Returns:
        datetime.datetime: A timezone-aware datetime object
    """
    return datetime.datetime.now(tz=datetime.timezone.utc)



"""
Utilities for interactively prompting the user for stuff at the console.
"""

from typing import List
from typing import TypeVar

_T = TypeVar("_T")


def get_user_choices(
    choices: List[_T],
    prompt_message: str = f"Choose from the following",
    allow_multiple: bool = True,
    allow_empty: bool = True,
    verify: bool = True,
) -> List[_T]:
    """Prompt user to select a subset of given choices interactively at console.

    Args:
        choices (List[_T]): The valid choices to be selected from. Each entry should have a valid __str__() definition.
        prompt_message (str, optional): The prompt to show the user when selecting. Defaults to f'Choose from the following'.
        allow_multiple (bool, optional): Allow the user to select more than 1 entry. Defaults to True.
        allow_empty (bool, optional): Allow the user to select no entries. Will return an empty list. Defaults to True.
        verify (bool, optional): Present the user with the choices they made and ask them to verify correctness. Defaults to True.

    Returns:
        List[_T]: The subset of "choices" selected by the user
    """
    error_message = ""
    selected_choices: List
    while True:
        # Give the user their options
        print(f"\n{prompt_message}:")
        for index, choice in enumerate(choices):
            print(f"  {(index + 1):3}. {choice}")
        if error_message != "":
            print(f"[{error_message}] ", end="")
        if allow_empty:
            print(f"[Empty selection: ok] ", end="")
        else:
            print(f"[Empty selection: NOT OK] ", end="")
        if allow_multiple:
            print(f"[Multiple selection: ok]")
        else:
            print(f"[Multiple selection: NOT OK]")

        # Parse user index choices
        user_input_raw = input(
            f'Enter choice(s), separated by spaces. Blank to select nothing, "all" to select everything: '
        )
        if user_input_raw.strip() == "":
            selected_indexes = []
        elif user_input_raw.strip().strip("\"'").casefold() == "all":
            selected_indexes = [x for x in range(len(choices))]
        else:
            try:
                selected_indexes = [int(x) - 1 for x in user_input_raw.split()]
            except Exception as exc:
                error_message = f"Error parsing user choice(s): {exc}. Please try again."
                continue

        # Make sure number of selections is valid
        if (not allow_empty) and len(selected_indexes) == 0:
            error_message = (
                f"You must select at least 1 option. You selected {len(selected_indexes)}. Please try again."
            )
            continue
        if (not allow_multiple) and len(selected_indexes) > 1:
            error_message = f"You must select exactly 1 option. You selected {len(selected_indexes)}. Please try again."
            continue

        # Remove duplicates and sort in ascending order
        selected_indexes = list(set(selected_indexes))
        selected_indexes.sort()

        # Convert the indexes to a choice list
        selected_choices: List[_T] = []
        try:
            for index in selected_indexes:
                if index < 0:
                    # User input "0" becomes index -1
                    # Which is not an index error, usually.
                    # But we don't want to allow it.
                    raise IndexError
                selected_choices.append(choices[index])
        except IndexError:
            error_message = f"You selected an invalid index. Please try again."
            continue

        # At this point, input is valid.
        # If we don't need to verify, we're done
        if not verify:
            break

        print(f"\nYou selected {len(selected_indexes)} choice(s):")
        for index in selected_indexes:
            print(f"  {(index + 1):3}. {choices[index]}")
        user_input_verify = input(f"Is this correct? (y/n): ").strip().strip("\"'").casefold()
        if user_input_verify == "y" or user_input_verify == "yes":
            break
        elif user_input_verify == "n" or user_input_verify == "no":
            error_message = f""
        else:
            error_message = f"Could not interpret yes/no input. Please try again."

    return selected_choices


def get_user_choice_single(
    choices: List[_T],
    prompt_message: str = f"Choose from the following",
    verify: bool = True,
) -> _T:
    return get_user_choices(
        choices=choices,
        prompt_message=prompt_message,
        allow_multiple=False,
        allow_empty=False,
        verify=verify,
    )[0]


if __name__ == "__main__":
    choices = "AAA BBB CCC DDD EEE".split()
    # choices = [Path(x).resolve().absolute() for x in os.listdir()][:5]

    result = get_user_choices(choices, allow_multiple=True, allow_empty=True, verify=True)
    print(result)
